from functools import reduce

import re

from vit import util
from vit.process import Command

class AutoComplete:

    def __init__(self, config, default_filters=None, extra_filters=None):
        self.default_filters = default_filters or ('column', 'project', 'tag')
        self.extra_filters = extra_filters or {}
        self.default_filter_config = {
            'column': {
                'suffixes': [':'],
            },
            'project': {
                'prefixes': ['project:'],
            },
            'tag': {
                'prefixes': ['+', '-'],
            },
        }
        self.config = config
        self.command = Command(self.config)
        for ac_type in self.default_filters:
            setattr(self, ac_type, [])
        for ac_type, items in list(self.extra_filters.items()):
            setattr(self, ac_type, items)
        self.reset()

    def refresh(self, filters=None):
        filters = filters or self.default_filters
        for ac_type in filters:
            setattr(self, ac_type, self.refresh_type(ac_type))

    def get_refresh_type_command(self, ac_type):
        command = [
            'task',
        ]
        if ac_type == 'project':
            command.extend([
                'rc.list.all.projects=yes',
                '_projects',
            ])
        else:
            command.extend([
                '_%ss' % ac_type
            ])
        return command

    def refresh_type(self, ac_type):
        returncode, stdout, stderr = self.command.run(self.get_refresh_type_command(ac_type), capture_output=True)
        if returncode == 0:
            items = list(filter(lambda x: True if x else False, stdout.split("\n")))
            if ac_type == 'project':
                items = self.create_project_entries(items)
            return items
        else:
            raise RuntimeError("Error running command '%s': %s" % (command, stderr))

    def create_project_entries(self, projects):
        def projects_reducer(projects_accum, project):
            def project_reducer(project_accum, part):
                project_accum.append(part)
                project_string = '.'.join(project_accum)
                if not project_string in projects_accum:
                    projects_accum.append(project_string)
                return project_accum
            reduce(project_reducer, project.split('.'), [])
            return projects_accum
        return reduce(projects_reducer, projects, [])

    def make_entries(self, filters, filter_config):
        entries = []
        for ac_type in filters:
            items = getattr(self, ac_type)
            include_unprefixed = filter_config[ac_type]['include_unprefixed'] if ac_type in filter_config and 'include_unprefixed' in filter_config[ac_type] else False
            type_prefixes = filter_config[ac_type]['prefixes'] if ac_type in filter_config and 'prefixes' in filter_config[ac_type] else []
            type_suffixes = filter_config[ac_type]['suffixes'] if ac_type in filter_config and 'suffixes' in filter_config[ac_type] else []
            if include_unprefixed:
                for item in items:
                    entries.append((ac_type, item))
            for prefix in type_prefixes:
                for item in items:
                    entries.append((ac_type, '%s%s' % (prefix, item)))
            for suffix in type_suffixes:
                for item in items:
                    entries.append((ac_type, '%s%s' % (item, suffix)))
        entries.sort()
        return entries

    def make_space_escape_regex(self, filters, filter_config):
        prefix_parts = []
        for ac_type in filters:
            items = getattr(self, ac_type)
            type_prefixes = filter_config[ac_type]['prefixes'] if ac_type in filter_config and 'prefixes' in filter_config[ac_type] else []
            for prefix in type_prefixes:
                prefix_parts.append(re.escape(prefix))
        prefix_or = "|".join(prefix_parts)
        return re.compile("^(%s).+[ ]+.+$" % prefix_or)

    def setup(self, text_callback, filters=None, filter_config=None):
        if self.is_setup:
            self.reset()
        self.text_callback = text_callback
        if not filters:
            filters = self.default_filters
        if not filter_config:
            filter_config = self.default_filter_config
        self.refresh()
        self.entries = self.make_entries(filters, filter_config)
        self.space_escape_regex = self.make_space_escape_regex(filters, filter_config)
        self.root_only_filters = list(filter(lambda f: True if f in filter_config and 'root_only' in filter_config[f] else False, filters))
        self.is_setup = True

    def teardown(self):
        self.is_setup = False
        self.entries = []
        self.root_only_filters = []
        self.callback = None
        self.deactivate()

    def reset(self):
        self.teardown()

    def activate(self, text, edit_pos, reverse=False):
        if not self.is_setup:
            return
        if self.activated:
            self.send_tabbed_text(text, edit_pos, reverse)
            return
        if self.can_tab(text, edit_pos):
            self.activated = True
            self.generate_tab_options(text, edit_pos)
            self.send_tabbed_text(text, edit_pos, reverse)

    def deactivate(self):
        self.activated = False
        self.idx = None
        self.tab_options = []
        self.root_search = False
        self.search_fragment = None
        self.prefix = None
        self.suffix = None
        self.partial = None

    def send_tabbed_text(self, text, edit_pos, reverse):
        tabbed_text, final_edit_pos = self.next_tab_item(text, reverse)
        self.text_callback(tabbed_text, final_edit_pos)

    def generate_tab_options(self, text, edit_pos):
        if self.root_search:
            if self.has_root_only_filters():
                self.tab_options = list(map(lambda e: e[1], filter(lambda e: True if e[0] in self.root_only_filters else False, self.entries)))
            else:
                self.tab_options = list(map(lambda e: e[1], self.entries))
        else:
            self.parse_text(text, edit_pos)
            exp = self.regexify(self.search_fragment)
            if self.has_root_only_filters():
                if self.search_fragment_is_root():
                    self.tab_options = list(map(lambda e: e[1], filter(lambda e: True if e[0] in self.root_only_filters and exp.match(e[1]) else False, self.entries)))
                else:
                    self.tab_options = list(map(lambda e: e[1], filter(lambda e: True if e[0] not in self.root_only_filters and exp.match(e[1]) else False, self.entries)))
            else:
                self.tab_options = list(map(lambda e: e[1], filter(lambda e: True if exp.match(e[1]) else False, self.entries)))

    def has_root_only_filters(self):
        return len(self.root_only_filters) > 0

    def search_fragment_is_root(self):
        return len(self.prefix_parts) == 0 or self.is_help_request()

    def regexify(self, string):
        return re.compile(re.escape(string))

    # TODO: This is way hacky, not sure of a cleaner way to handle
    # multi-spaced search terms, of which help is the only one now.
    def is_help_request(self):
        return self.prefix_parts[0] in ['help']

    def add_space_escaping(self, text):
        if self.space_escape_regex.match(text):
            parts = text.split(':', 1)
            if len(parts) > 1:
                return "%s:'%s'" % (parts[0], parts[1])
            else:
                return "'%s'" % text
        else:
            return text

    def remove_space_escaping(self, text):
        return text.replace('\\ ', ' ')

    def parse_text(self, text, edit_pos):
        full_prefix = text[:edit_pos]
        self.prefix_parts = list(map(self.add_space_escaping, util.string_to_args_on_whitespace(full_prefix)))
        if not self.prefix_parts:
            self.search_fragment = self.prefix = full_prefix
            self.suffix = text[(edit_pos + 1):]
        elif self.is_help_request():
            self.search_fragment = full_prefix
            self.prefix = self.suffix = ''
        else:
            self.search_fragment = self.prefix_parts.pop()
            self.prefix = ' '.join(self.prefix_parts)
            self.suffix = text[(edit_pos + 1):]
        self.search_fragment = self.remove_space_escaping(self.search_fragment)

    def can_tab(self, text, edit_pos):
        if edit_pos == 0:
            if text == '':
                self.root_search = True
                return True
            return False
        previous_pos = edit_pos - 1
        next_pos = edit_pos + 1
        return text[edit_pos:next_pos] in (' ', '') and text[previous_pos:edit_pos] not in (' ', '')

    def assemble(self, tab_option, solo_match=False):
        if not tab_option.endswith(":"):
            tab_option = self.add_space_escaping(tab_option)
            if solo_match:
                tab_option += ' '
        parts = [self.prefix, tab_option, self.suffix]
        tabbed_text = ' '.join(filter(lambda p: True if p else False, parts))
        parts.pop()
        edit_pos_parts = ' '.join(filter(lambda p: True if p else False, parts))
        edit_pos_final = len(edit_pos_parts)
        return tabbed_text, edit_pos_final

    def partial_match(self):
        if self.partial:
            return
        ref_item = self.tab_options[0]
        ref_item_length = len(ref_item)
        tab_options_length = len(self.tab_options)
        pos = len(self.search_fragment)
        self.partial = self.search_fragment
        while pos < ref_item_length:
            pos += 1
            exp = self.regexify(ref_item[:pos])
            ref_result = list(filter(lambda o: True if exp.match(o) else False, self.tab_options))
            if len(ref_result) == tab_options_length:
                self.partial = ref_item[:pos]
            else:
                break
        return self.partial != self.search_fragment

    def initial_idx(self, reverse):
        return len(self.tab_options) - 1 if reverse else 0

    def increment_index(self, reverse):
        if self.idx == None:
            self.idx = self.initial_idx(reverse)
        else:
            if reverse:
                self.idx = self.idx - 1 if self.idx > 0 else len(self.tab_options) - 1
            else:
                self.idx = self.idx + 1 if self.idx < len(self.tab_options) - 1 else 0

    def next_tab_item(self, text, reverse):
        tabbed_text = ''
        edit_pos = None
        if self.root_search:
            self.increment_index(reverse)
            tabbed_text = self.tab_options[self.idx]
        else:
            if len(self.tab_options) == 0:
                tabbed_text = text
            elif len(self.tab_options) == 1:
                tabbed_text, edit_pos = self.assemble(self.tab_options[0], solo_match=True)
            else:
                if self.partial_match():
                    tabbed_text, edit_pos = self.assemble(self.partial)
                else:
                    if self.idx == None and self.partial == self.tab_options[self.initial_idx(reverse)]:
                        self.increment_index(reverse)
                    self.increment_index(reverse)
                    tabbed_text, edit_pos = self.assemble(self.tab_options[self.idx])
        return tabbed_text, edit_pos




from vit.formatter import Number
from vit.util import unicode_len

class UdaNumeric(Number):
    def format(self, number, task):
        if number is None:
            return self.markup_none(self.colorize())
        number = str(number)
        return (unicode_len(number), self.markup_element(number))
    def colorize(self, number=None):
        return self.colorizer.uda_numeric(self.column, number)



import re
BRIGHT_REGEX = re.compile('.*bright.*')

def task_256_to_urwid_256():
    manual_map = {
        'red': 'dark red',
        'green': 'dark green',
        'blue': 'dark blue',
        'cyan': 'dark cyan',
        'magenta': 'dark magenta',
        'gray': 'light gray',
        'yellow': 'brown',
        'color0': 'black',
        'color1': 'dark red',
        'color2': 'dark green',
        'color3': 'brown',
        'color4': 'dark blue',
        'color5': 'dark magenta',
        'color6': 'dark cyan',
        'color7': 'light gray',
        'color8': 'dark gray',
        'color9': 'light red',
        'color10': 'light green',
        'color11': 'yellow',
        'color12': 'light blue',
        'color13': 'light magenta',
        'color14': 'light cyan',
        'color15': 'white',
    }
    manual_map.update(task_color_gray_to_g())
    manual_map.update(task_color_to_h())
    manual_map.update(task_rgb_to_h())
    return manual_map

def task_bright_to_color(color_string):
    color_map = {
        'bright black': 'color8',
        'bright red': 'color9',
        'bright green': 'color10',
        'bright yellow': 'color11',
        'bright blue': 'color12',
        'bright magenta': 'color13',
        'bright cyan': 'color14',
        'bright white': 'color15',
    }
    if BRIGHT_REGEX.match(color_string):
        for bright_color in color_map:
            color_string = color_string.replace(bright_color, color_map[bright_color])
    return color_string

def task_color_gray_to_g():
    color_map = {}
    for i in range(0, 24):
        gray_key = 'gray%d' % i
        color_key = 'color%d' % (i + 232)
        # NOTE: This is an approximation of the conversion, close enough!
        value = 'g%d' % (i * 4)
        color_map[gray_key] = value
        color_map[color_key] = value
    return color_map

def task_color_to_h():
    color_map = {}
    for i in range(16, 232):
        key = 'color%d' % i
        value = 'h%d' % i
        color_map[key] = value
    return color_map

def task_rgb_to_h():
    index_to_hex = [
        '0',
        '6',
        '8',
        'a',
        'd',
        'f',
    ]
    color_map = {}
    count = 0
    for r in range(0, 6):
        for g in range(0, 6):
            for b in range(0, 6):
                key = 'rgb%d%d%d' % (r, g, b)
                value = '#%s%s%s' % (index_to_hex[r], index_to_hex[g], index_to_hex[b])
                color_map[key] = value
                count += 1
    return color_map



import unicodedata
from vit.formatter import Marker
from vit.util import unicode_len

class Markers(Marker):
    def format(self, _, task):
        text_markup = []
        width = 0
        if self.mark_tags:
            width, text_markup = self.format_tags(width, text_markup, task['tags'])
        if self.mark_project:
            width, text_markup = self.format_project(width, text_markup, task['project'])
        if self.mark_due and task['due']:
            width, text_markup = self.format_due(width, text_markup, task['due'], task)
        if self.mark_status:
            width, text_markup = self.format_status(width, text_markup, task['status'])
        if self.mark_depends and self.filter_by_blocking_task_uuids(task['depends']):
            width, text_markup = self.format_blocked(width, text_markup, task['depends'])
        if self.mark_start and task['start']:
            width, text_markup = self.format_active(width, text_markup, task['start'], task)
        if self.mark_recur and task['recur']:
            width, text_markup = self.format_recurring(width, text_markup, task['recur'])
        if self.mark_scheduled and task['scheduled']:
            width, text_markup = self.format_scheduled(width, text_markup, task['scheduled'], task)
        if self.mark_until and task['until']:
            width, text_markup = self.format_until(width, text_markup, task['until'], task)
        for uda_name, uda_type in self.udas.items():
            if getattr(self, 'mark_%s' % uda_name):
                width, text_markup = self.format_uda(width, text_markup, uda_name, uda_type, task[uda_name])
        if task['uuid'] in self.blocking_task_uuids:
            width, text_markup = self.format_blocking(width, text_markup)
        return (width, '' if width == 0 else text_markup)

    def color_required(self, color):
        return self.require_color and not color

    def add_label(self, color, label, width, text_markup):
        if self.color_required(color) or not label:
            return width, text_markup
        width += unicode_len(label)
        text_markup += [(color, label)]
        return width, text_markup

    def format_tags(self, width, text_markup, tags):
        if not tags:
            color = self.colorizer.tag_none()
            label = self.labels['tag.none.label']
            return self.add_label(color, label, width, text_markup)
        elif len(tags) == 1:
            tag = list(tags)[0]
            color = self.colorizer.tag(tag)
            custom_label = 'tag.%s.label' % tag
            label = self.labels['tag.label'] if not custom_label in self.labels else self.labels[custom_label]
            return self.add_label(color, label, width, text_markup)
        else:
            color = self.colorizer.tag('')
            label = self.labels['tag.label']
            return self.add_label(color, label, width, text_markup)

    def format_project(self, width, text_markup, project):
        if not project:
            color = self.colorizer.project_none()
            label = self.labels['project.none.label']
            return self.add_label(color, label, width, text_markup)
        else:
            color = self.colorizer.project(project)
            custom_label = 'project.%s.label' % project
            label = self.labels['project.label'] if not custom_label in self.labels else self.labels[custom_label]
            return self.add_label(color, label, width, text_markup)

    def format_due(self, width, text_markup, due, task):
        due_state = self.formatter.get_due_state(due, task)
        if due_state:
            color = self.colorizer.due(due_state)
            label = self.labels['%s.label' % due_state]
            return self.add_label(color, label, width, text_markup)
        return width, text_markup

    def format_uda(self, width, text_markup, uda_name, uda_type, value):
        if not value:
            color = self.colorizer.uda_none(uda_name)
            label = self.labels['uda.%s.none.label' % uda_name]
            return self.add_label(color, label, width, text_markup)
        else:
            color = getattr(self.colorizer, 'uda_%s' % uda_type)(uda_name, value)
            custom_label = 'uda.%s.label' % uda_name
            label = self.labels['uda.label'] if not custom_label in self.labels else self.labels[custom_label]
            return self.add_label(color, label, width, text_markup)

    def format_blocking(self, width, text_markup):
        color = self.colorizer.blocking()
        label = self.labels['blocking.label']
        return self.add_label(color, label, width, text_markup)

    def format_status(self, width, text_markup, status):
        if status == 'completed' or status == 'deleted':
            color = self.colorizer.status(status)
            label = self.labels['%s.label' % status]
            return self.add_label(color, label, width, text_markup)
        return width, text_markup

    def format_blocked(self, width, text_markup, depends):
        color = self.colorizer.blocked(depends)
        label = self.labels['blocked.label']
        return self.add_label(color, label, width, text_markup)

    def format_active(self, width, text_markup, start, task):
        active = self.formatter.get_active_state(start, task)
        if active:
            color = self.colorizer.active(active)
            label = self.labels['active.label']
            return self.add_label(color, label, width, text_markup)
        return width, text_markup

    def format_recurring(self, width, text_markup, recur):
        color = self.colorizer.recurring(recur)
        label = self.labels['recurring.label']
        return self.add_label(color, label, width, text_markup)

    def format_scheduled(self, width, text_markup, scheduled, task):
        scheduled = self.formatter.get_scheduled_state(scheduled, task)
        if scheduled:
            color = self.colorizer.scheduled(scheduled)
            label = self.labels['scheduled.label']
            return self.add_label(color, label, width, text_markup)
        return width, text_markup


    def format_until(self, width, text_markup, until, task):
        until = self.formatter.get_until_state(until, task)
        if until:
            color = self.colorizer.until(until)
            label = self.labels['until.label']
            return self.add_label(color, label, width, text_markup)
        return width, text_markup



from vit.formatter.start import Start

class StartActive(Start):
    def format_datetime(self, start, task):
        return self.formatter.indicator_active if self.get_active_state(start, task) else ''



from vit.formatter.tags import Tags

class TagsIndicator(Tags):
    def format(self, tags, task):
        if not tags:
            return self.markup_none(self.colorizer.tag_none())
        else:
            return (1, (self.colorizer.tag(''), self.formatter.indicator_tag))



import math
from datetime import datetime
try:
    from zoneinfo import ZoneInfo
except ImportError:
    from backports.zoneinfo import ZoneInfo

from vit.util import unicode_len

TIME_UNIT_MAP = {
    'seconds': {
        'label': 's',
        'divisor': 1,
        'threshold': 60,
     },
    'minutes': {
        'label': 'm',
        'divisor': 60,
        'threshold': 3600,
     },
    'hours': {
        'label': 'h',
        'divisor': 3600,
        'threshold': 86400,
     },
    'days': {
        'label': 'd',
        'divisor': 86400,
        'threshold': 86400 * 14,
     },
    'weeks': {
        'label': 'w',
        'divisor': 86400 * 7,
        'threshold': 86400 * 90,
     },
    'months': {
        'label': 'mo',
        'divisor': 86400 * 30,
        'threshold': 86400 * 365,
     },
    'years': {
        'label': 'y',
        'divisor': 86400 * 365,
     },
}

class Formatter:
    def __init__(self, column, report, formatter_base, blocking_task_uuids, **kwargs):
        self.column = column
        self.report = report
        self.formatter = formatter_base
        self.colorizer = self.formatter.task_colorizer
        self.blocking_task_uuids = blocking_task_uuids

    def format(self, obj, task):
        if not obj:
            return self.empty()
        obj = str(obj)
        return (unicode_len(obj), self.markup_element(obj))

    def empty(self):
        return (0, '')

    def markup_element(self, obj):
        return (self.colorize(obj), obj)

    def markup_none(self, color):
        if color:
            return (unicode_len(self.formatter.none_label), (color, self.formatter.none_label))
        else:
            return self.empty()

    def colorize(self, obj):
        return None

    def filter_by_blocking_task_uuids(self, depends):
        return [ task for task in depends if task['uuid'] in self.blocking_task_uuids ]

class Marker(Formatter):
    def __init__(self, report, defaults, report_marker_columns, blocking_task_uuids):
        super().__init__(None, report, defaults, blocking_task_uuids)
        self.columns = report_marker_columns
        self.labels = self.formatter.markers.labels
        self.udas = self.formatter.markers.udas
        self.require_color = self.formatter.markers.require_color
        self.set_column_attrs()

    def set_column_attrs(self):
        any(setattr(self, 'mark_%s' % column, column in self.columns) for column in self.formatter.markers.markable_columns)

class Number(Formatter):
    pass

class String(Formatter):
    pass

class Duration(Formatter):
    def format(self, obj, task):
        if not obj:
            return self.empty()
        formatted_duration = self.format_duration(obj)
        return (unicode_len(formatted_duration), self.markup_element(obj, formatted_duration))

    def format_duration(self, obj):
        return obj

    def markup_element(self, obj, formatted_duration):
        return (self.colorize(obj), formatted_duration)

class DateTime(Formatter):
    def __init__(self, column, report, defaults, blocking_task_uuids, **kwargs):
        self.custom_formatter = None if not 'custom_formatter' in kwargs else kwargs['custom_formatter']
        super().__init__(column, report, defaults, blocking_task_uuids)

    def format(self, dt, task):
        if not dt:
            return self.empty()
        formatted_date = self.format_datetime(dt, task)
        return (unicode_len(formatted_date), self.markup_element(dt, formatted_date, task))

    def format_datetime(self, dt, task):
        return dt.strftime(self.custom_formatter or self.formatter.report)

    def markup_element(self, dt, formatted_date, task):
        return (self.colorize(dt, task), formatted_date)

    def colorize(self, dt, task):
        return None

    def format_duration_vague(self, seconds):
        test = seconds
        sign = ''
        unit = 'seconds'
        if seconds < 0:
            test = -seconds
            sign = '-'
        if test <= TIME_UNIT_MAP['seconds']['threshold']:
            # Handled by defaults
            pass
        elif test <= TIME_UNIT_MAP['minutes']['threshold']:
            unit = 'minutes'
        elif test <= TIME_UNIT_MAP['hours']['threshold']:
            unit = 'hours'
        elif test <= TIME_UNIT_MAP['days']['threshold']:
            unit = 'days'
        elif test <= TIME_UNIT_MAP['weeks']['threshold']:
            unit = 'weeks'
        elif test <= TIME_UNIT_MAP['months']['threshold']:
            unit = 'months'
        else:
            unit = 'years'
        age = test // TIME_UNIT_MAP[unit]['divisor']
        return '%s%d%s' % (sign, age, TIME_UNIT_MAP[unit]['label'])

    def age(self, dt):
        if dt == None:
            return ''
        now = datetime.now().astimezone()
        seconds = (now - dt).total_seconds()
        return self.format_duration_vague(seconds)

    def countdown(self, dt):
        if dt == None:
            return ''
        now = datetime.now().astimezone()
        if dt < now:
            return ''
        seconds = (dt - now).total_seconds()
        return self.format_duration_vague(seconds)

    def relative(self, dt):
        if dt == None:
            return ''
        now = datetime.now().astimezone()
        seconds = (dt - now).total_seconds()
        return self.format_duration_vague(seconds)

    def remaining(self, dt):
        if dt == None:
            return ''
        now = datetime.now().astimezone()
        if dt < now:
            return ''
        seconds = (dt - now).total_seconds()
        return self.format_duration_vague(seconds)

    def epoch(self, dt):
        if dt == None:
            return ''
        return str(round((dt - self.formatter.epoch_datetime).total_seconds()))

    # Taken from https://github.com/dannyzed/julian
    def julian(self, dt):
        if dt == None:
            return ''
        a = math.floor((14-dt.month)/12)
        y = dt.year + 4800 - a
        m = dt.month + 12*a - 3
        jdn = dt.day + math.floor((153*m + 2)/5) + 365*y + math.floor(y/4) - math.floor(y/100) + math.floor(y/400) - 32045
        jd = jdn + (dt.hour - 12) / 24 + dt.minute / 1440 + dt.second / 86400 + dt.microsecond / 86400000000
        return str(jd)

    def iso(self, dt):
        if dt == None:
            return ''
        dt = dt.replace(tzinfo=ZoneInfo('UTC'))
        return dt.isoformat()

class List(Formatter):

    def format(self, obj, task):
        if not obj:
            return self.empty()
        formatted = self.format_list(obj, task)
        return (unicode_len(formatted), self.markup_element(obj, formatted))

    def markup_element(self, obj, formatted):
        return (self.colorize(obj), formatted)

    def format_list(self, obj, task):
        return ','.join(obj) if obj else ''



from vit.formatter import String
from vit.util import unicode_len

class UdaString(String):
    def format(self, string, task):
        if not string:
            return self.markup_none(self.colorize())
        return (unicode_len(string), self.markup_element(string))
    def colorize(self, string=None):
        return self.colorizer.uda_string(self.column, string)



import glob
import sys
import argparse

from vit import version

parser = argparse.ArgumentParser(
    description="VIT (Visual Interactive Taskwarrior)",
    usage='%(prog)s [options] [report] [filters]',
    formatter_class=argparse.RawTextHelpFormatter,
    allow_abbrev=False,
    epilog="""
VIT (Visual Interactive Taskwarrior) is a lightweight, curses-based front end for
Taskwarrior that provides a convenient way to quickly navigate and process tasks.
VIT allows you to interact with tasks in a Vi-intuitive way.

A goal of VIT is to allow you to customize the way in which you use Taskwarrior's
core commands as well as to provide a framework for easily dispatching external
commands.

While VIT is running, type :help followed by enter to review basic command/navigation actions.

See https://github.com/vit-project/vit for more information.

"""
)

parser.add_argument('-v', '--version',
    action='version',
    version=version.VIT,
)
parser.add_argument('--list-actions',
      dest="list_actions",
      default=False,
      action="store_true",
      help="list all available actions",
)
parser.add_argument('--list-pids',
      dest="list_pids",
      default=False,
      action="store_true",
      help="list all pids found in pid_dir, if configured",
)

def parse_options():
    options, filters = parser.parse_known_args()
    if options.list_actions:
        list_actions()
        sys.exit(0)
    elif options.list_pids:
        ret = list_pids()
        sys.exit(ret)
    return options, filters

def format_dictionary_list(item, description):
    print("%s:" % item)
    print("\t%s\n" % description)

def list_actions():
    from vit.registry import ActionRegistry
    from vit.actions import Actions
    action_registry = ActionRegistry()
    actions = Actions(action_registry)
    actions.register()
    any(format_dictionary_list(action, data['description']) for action, data in actions.get().items())

def _get_pids_from_pid_dir(pid_dir):
        filepaths = glob.glob("%s/*.pid" % pid_dir)
        pids = []
        for filepath in filepaths:
            try:
                with open(filepath, 'r') as f:
                    pids.append(f.read())
            except IOError:
                pass
        return pids

def list_pids():
    from vit.loader import Loader
    from vit.config_parser import ConfigParser
    from vit.pid_manager import PidManager
    loader = Loader()
    config = ConfigParser(loader)
    pid_manager = PidManager(config)
    if pid_manager.pid_dir:
        pids = _get_pids_from_pid_dir(pid_manager.pid_dir)
        print("\n".join(pids))
        return 0
    else:
        print("ERROR: No pid_dir configured")
        return 1



from vit import uda

MARKABLE_COLUMNS = [
    'depends',
    'description',
    'due',
    'project',
    'recur',
    'scheduled',
    'start',
    'status',
    'tags',
    'until',
]
LABEL_DEFAULTS = {
    'active.label': '(A)',
    'blocked.label': '(BD)',
    'blocking.label': '(BG)',
    'completed.label': '(C)',
    'deleted.label': '(X)',
    'due.label': '(D)',
    'due.today.label': '(DT)',
    'keyword.label': '(K)',
    'overdue.label': '(OD)',
    'project.label': '(P)',
    'project.none.label': '',
    'recurring.label': '(R)',
    'scheduled.label': '(S)',
    'tag.label': '(T)',
    'tag.none.label': '',
    'uda.label': '',
    'uda.priority.label': '(PR)',
    'until.label': '(U)',
}

class Markers:
    def __init__(self, config, task_config):
        self.config = config
        self.task_config = task_config
        self.enabled = self.config.get('marker', 'enabled')
        if self.enabled:
            self.udas = uda.get_configured(self.task_config)
            self.markable_columns = MARKABLE_COLUMNS + list(self.udas.keys())
            self.configured_columns = self.config.get('marker', 'columns')
            self.set_columns()
            self.header_label = self.config.get('marker', 'header_label')
            self.require_color = self.config.get('marker', 'require_color')
            self.include_subprojects = self.config.get('marker', 'include_subprojects')
            self.compose_labels()
            self.set_none_label_attributes()

    def set_columns(self):
        self.columns = self.markable_columns if self.configured_columns == 'all' else self.configured_columns.split(',')

    def compose_labels(self):
        self.labels = LABEL_DEFAULTS.copy()
        for uda in list(self.udas.keys()):
            self.labels['uda.%s.none.label' % uda] = ''
        for key, value in self.config.items('marker'):
            if key not in self.config.defaults['marker']:
                self.labels[key] = value
        if self.include_subprojects:
            self.add_project_children()

    def set_none_label_attributes(self):
        for label_type in ['project', 'tag', 'uda.priority']:
            label = '%s.none.label' % label_type
            attr = '%s_has_none_marker' % label_type.replace('.', '_')
            has_marker = self.labels[label] != ''
            setattr(self, attr, has_marker)

    def add_project_children(self):
        project_prefix = 'project.'
        label_suffix = '.label'
        for label, marker in self.get_project_labels():
            for entry in self.task_config.projects:
                new_label = '%s%s' % (project_prefix, entry)
                if not self.has_label(new_label) and new_label.startswith(label[:-len(label_suffix)]):
                    self.labels['%s%s' % (new_label, label_suffix)] = marker

    def get_project_labels(self):
        return sorted([(label, marker) for label, marker in self.labels.items() if self.is_project_label(label)], reverse=True)

    def is_project_label(self, label):
        return label.startswith('project.') and label not in ['project.label', 'project.none.label']

    def has_label(self, label):
        return label in self.labels



import os
import errno

class PidManager:
    """Simple process ID manager.
    """
    def __init__(self, config):
        self.config = config
        self.uid = os.getuid()
        self.pid = os.getpid()
        self._format_pid_dir()
        self._make_pid_filepath()

    def setup(self):
        if self.pid_dir:
            self._create_pid_dir()
            self._write_pid_file()

    def teardown(self):
        if self.pid_dir:
            try:
                os.remove(self.pid_file)
            # TODO: This needs a little more work to skip errors when no PID file
            #       exists.
            #except OSError as e:
            #    if e.errno != errno.ENOENT:
            #        raise OSError("could not remove pid file %s" % self.pid_file)
            except:
                pass

    def _format_pid_dir(self):
        config_pid_dir = self.config.get('vit', 'pid_dir')
        self.pid_dir = config_pid_dir.replace("$UID", str(self.uid))

    def _make_pid_filepath(self):
        self.pid_file = "%s/%s.pid" % (self.pid_dir, self.pid)

    def _create_pid_dir(self):
        try:
            os.makedirs(self.pid_dir, exist_ok=True)
        except OSError:
            raise OSError("could not create pid_dir %s" % self.pid_dir)

    def _write_pid_file(self):
        try:
            with open(self.pid_file, "w") as f:
                f.write(str(self.pid))
        except IOError:
            raise IOError("could not write pid file %s" % self.pid_file)



import urwid
from vit.readline import Readline

class CommandBar(urwid.Edit):
    """Custom urwid.Edit class for the command bar.
    """
    def __init__(self, **kwargs):
        self.event = kwargs.pop('event')
        self.autocomplete = kwargs.pop('autocomplete')
        self.abort_backspace = kwargs.pop('abort_backspace')
        self.metadata = None
        self.history = CommandBarHistory()
        self.readline = Readline(self)
        return super().__init__(**kwargs)

    def keypress(self, size, key):
        """Overrides Edit.keypress method.
        """
        if key not in ('tab', 'shift tab'):
            self.autocomplete.deactivate()
        if 'choices' in self.metadata:
            self.quit({'choice': self.metadata['choices'].get(key)})
            return None
        elif key in ('up',):
            self.readline.keypress('ctrl p')
            return None
        elif key in ('down',):
            self.readline.keypress('ctrl n')
            return None
        elif key in ('enter', 'esc'):
            text = self.get_edit_text().strip()
            if text and key == 'enter':
                self.history.add(self.metadata['history'], text)
            self.quit({'key': key, 'text': text})
            return None
        elif key in ('tab', 'shift tab'):
            if self.is_autocomplete_op():
                text = self.get_edit_text()
                kwargs = {}
                if key in ('shift tab',):
                    kwargs['reverse'] = True
                self.autocomplete.activate(text, self.edit_pos, **kwargs)
            return None
        elif key in self.readline.keys():
            return self.readline.keypress(key)
        elif self.is_aborting_backspace(key):
            self.quit({'key': key})
            return None
        return super().keypress(size, key)

    def is_aborting_backspace(self, key):
        return key == 'backspace' and self.abort_backspace and not self.get_edit_text()

    def is_autocomplete_op(self):
        return self.metadata['op'] not in ['search-forward', 'search-reverse']

    def set_edit_text(self, text, edit_pos=None):
        ret = super().set_edit_text(text)
        if not edit_pos:
            edit_pos = len(text)
        self.set_edit_pos(edit_pos)
        return ret

    def set_command_prompt(self, caption, edit_text=None):
        self.set_caption(caption)
        if edit_text is not None:
            self.set_edit_text(edit_text)

    def activate(self, caption, metadata, edit_text=None):
        self.set_metadata(metadata)
        self.set_command_prompt(caption, edit_text)

    def deactivate(self):
        self.set_command_prompt('', '')
        self.history.cleanup(self.metadata['op'])
        self.set_metadata(None)

    def quit(self, metadata_args={}):
        data = {'metadata': self.get_metadata(), **metadata_args}
        self.deactivate()
        self.event.emit('command-bar:keypress', data)  # remove focus from command bar

    def get_metadata(self):
        return self.metadata.copy() if self.metadata else None

    def prepare_metadata(self, metadata):
        if metadata:
            if 'history' not in metadata:
                metadata['history'] = metadata['op']
        return metadata

    def set_metadata(self, metadata):
        self.metadata = self.prepare_metadata(metadata)

    def set_edit_text_callback(self):
        return self.set_edit_text

class CommandBarHistory:
    """Holds command-specific history for the command bar.
    """
    def __init__(self):
        self.commands = {}
        self.scrolling = False

    def add(self, command, text):
        if not self.exists(command):
            self.commands[command] = {'items': ['']}
        self.commands[command]['items'].insert(len(self.get_items(command)) - 1, text)
        self.set_scrolling(command, False)

    def previous(self, command):
        if self.exists(command):
            if not self.scrolling:
                self.set_scrolling(command, True)
                return self.current(command)
            elif self.get_idx(command) > 0:
                self.move_idx(command, -1)
                return self.current(command)
        return False

    def next(self, command):
        if self.exists(command) and self.scrolling and self.get_idx(command) < self.last_idx(command):
            self.move_idx(command, 1)
            return self.current(command)
        return False

    def cleanup(self, command):
        self.set_scrolling(command, False)

    def get_items(self, command):
        return self.commands[command]['items']

    def get_idx(self, command):
        return self.commands[command]['idx']

    def set_idx(self, command, idx):
        self.commands[command]['idx'] = idx

    def set_scrolling(self, command, scroll):
        if self.exists(command):
            # Don't count the ending empty string when setting the initial
            # index for scrolling to start.
            self.set_idx(command, self.last_idx(command) - 1)
        self.scrolling = scroll

    def move_idx(self, command, increment):
        self.set_idx(command, self.get_idx(command) + increment)

    def exists(self, command):
        return command in self.commands

    def current(self, command):
        return self.get_items(command)[self.get_idx(command)] if self.exists(command) else False

    def last_idx(self, command):
        return len(self.get_items(command)) - 1 if self.exists(command) else None




DEFAULT_BATCH_SIZE=100

class ListBatchError(Exception):
    pass

class ListBatcher:
    def __init__(self, batch_from, batch_to, batch_to_formatter=None, default_batch_size=DEFAULT_BATCH_SIZE):
        self.batch_from = batch_from
        self.batch_to = batch_to
        self.batch_to_formatter = batch_to_formatter
        self.default_batch_size = default_batch_size
        self.last_position = 0
        self.batching_complete = False

    def add(self, batch_size=None):
        if self.batching_complete:
            return True
        batch_size = self.get_batch_size(batch_size)
        self.load_batch(batch_size)
        if self.is_batching_complete():
            self.batching_complete = True
        return self.batching_complete

    def get_last_position(self):
        return self.last_position

    def load_batch(self, batch_size):
        new_position = self.last_position + batch_size
        partial = self.batch_from[self.last_position:new_position]
        if self.batch_to_formatter:
            partial = self.batch_to_formatter(partial, self.last_position)
        self.batch_to += partial
        self.last_position = new_position

    def is_batching_complete(self):
        return self.last_position >= len(self.batch_from)

    def batch_remainder(self):
        return len(self.batch_from) - self.last_position

    def get_batch_size(self, batch_size):
        remainder = self.batch_remainder()
        if batch_size == 0:
            return remainder
        else:
            if batch_size is None:
                batch_size = self.default_batch_size
            return remainder if batch_size > remainder else batch_size



from vit.formatter import DateTime

class Start(DateTime):
    def get_active_state(self, start, task):
        return self.formatter.get_active_state(start, task)

    def colorize(self, start, task):
        return self.colorizer.active(self.get_active_state(start, task))



import pprint
import tempfile

LOG_FILE = '%s/%s' % (tempfile.gettempdir(), 'vit-debug.log')

pp = pprint.PrettyPrinter()
pf = pprint.PrettyPrinter(stream=open(LOG_FILE,'w'))

def console(*args, **kwargs):
    pp.pprint(*args, **kwargs)

def file(*args, **kwargs):
    pf.pprint(*args, **kwargs)



import os
import sys
import curses
import shlex

from urwid.util import calc_width

curses.setupterm()
e3_seq = curses.tigetstr('E3') or b''
clear_screen_seq = curses.tigetstr('clear') or b''

def clear_screen():
    os.write(sys.stdout.fileno(), e3_seq + clear_screen_seq)

def string_to_args(string):
    try:
        return shlex.split(string)
    except ValueError:
        return []

def string_to_args_on_whitespace(string):
    try:
        lex = shlex.shlex(string)
        lex.whitespace_split = True
        return list(lex)
    except ValueError:
        return []

def is_mouse_event(key):
    return not isinstance(key, str)

def uuid_short(uuid):
    return uuid[0:8]

def task_id_or_uuid_short(task):
    return task['id'] or uuid_short(task['uuid'])

def task_pending(task):
    return task['status'] == 'pending'

def task_completed(task):
    return task['status'] == 'completed' or task['status'] == 'deleted'

def project_get_subproject_and_parents(project):
    parts = project.split('.')
    subproject = parts.pop()
    parents = parts if len(parts) > 0 else None
    return subproject, parents

def project_get_root(project):
    return project.split('.')[0] if project else None

def file_to_class_name(file_name):
    words = file_name.split('_')
    return ''.join((w.capitalize() for w in words))

def file_readable(filepath):
    return os.path.isfile(filepath) and os.access(filepath, os.R_OK)

def unicode_len(string):
    return calc_width(string, 0, len(string))
